-- ###########################################################################################
-- ### Lotto - Sessionversion ################################################################
-- ###########################################################################################
-- ### Anlage der Datenbank                                                                ###
-- ### ----------------------------------------------------------------------------------- ###
-- ### Dieses Skript erklaert den Umgang mit dem Zufall im SQl Server.                     ###
-- ###                                                                                     ###
-- ### ----------------------------------------------------------------------------------- ###
-- ### Sicherheitshinweis:                                                                 ###
-- ###      Ueber diese Befehlssammlung werden Datenbankobjekte angelegt, geaendert oder   ###
-- ###      geloescht. Auch koennen Inhalte hinzugefuegt, manipuliert oder entfernt        ###
-- ###      werden. In produktiven Umgebungen darf dieses Skript NICHT eingesetzt werden,  ###
-- ###      um versehentliche Auswirkungen auf sonstige Strukturen auszuschliessen.        ###
-- ###                                                                                     ###
-- ### Erstellung:                                                                         ###
-- ###      Torsten Ahlemeyer fuer arelium GmbH, www.arelium.de                            ###
-- ### ----------------------------------------------------------------------------------- ###
-- ### Aenderungsnachweis:                                                                 ###
-- ###     1.00.0	2022-07-26	Torsten Ahlemeyer                                          ###
-- ###              Initiale Erstellung mit Default-Werten                                 ###
-- ###########################################################################################




-- ########################################################
-- ### Variante 1: Eine einzelne Zufallszahl per RAND() ###
-- ########################################################


-- 1. eine Zufallszahl ausgeben
-- Zufallszahlen liegen per Definition stets zwischen 0 und 1
SELECT RAND();
GO



-- 2. zu Testzwecken kann man ein SEED angeben und so den Zufall wiederholbar machen
DECLARE @Seed			AS INTEGER
SET @Seed				= 13
SELECT RAND(@Seed);
GO



-- 3. Zufallszahl in einem bestimmten Range ausgeben
-- die Untergrenze ist inklusiv, die Obergrenze ist exklusiv!
DECLARE @Untergrenze	AS INTEGER
DECLARE @Obergrenze		AS INTEGER
SET @Untergrenze		= 1
SET @Obergrenze			= 50
SELECT RAND()*(@Obergrenze - @Untergrenze) + @Untergrenze;
GO



-- 4. Wenn Ganzzahlen bentigt werden, ist dieser Ausdruck zu konvertieren
DECLARE @Untergrenze	AS INTEGER
DECLARE @Obergrenze		AS INTEGER
SET @Untergrenze		= 1
SET @Obergrenze			= 50
SELECT CONVERT(INTEGER, RAND()*(@Obergrenze - @Untergrenze) + @Untergrenze)
SELECT FLOOR(RAND()*(@Obergrenze - @Untergrenze) + @Untergrenze)
-- FLOOR: Gibt die grte ganze Zahl zurck, die kleiner oder gleich dem angegebenen numerischen Ausdruck ist.
GO





-- 5a. Es bietet sich an, diesen Quellcode in eine Funktion mit Parametern auszulagern.
-- *** VORSICHT: So nicht! ***
CREATE OR ALTER FUNCTION [Lotto].[fncZufallszahl]
(
	  @Untergrenze		AS INTEGER
	, @Obergrenze		AS INTEGER
)
RETURNS INTEGER
AS
BEGIN
	DECLARE @Rueckgabewert		AS INTEGER
	SELECT @Rueckgabewert		= FLOOR(RAND() * (@Obergrenze - @Untergrenze) + @Untergrenze)
	RETURN @Rueckgabewert
END
GO



DECLARE @Untergrenze	AS INTEGER
DECLARE @Obergrenze		AS INTEGER
SET @Untergrenze		= 1
SET @Obergrenze			= 50

SELECT [Lotto].[fncZufallszahl] (@Untergrenze, @Obergrenze)
GO



-- 5b. Es bietet sich an, diesen Quellcode in eine Funktion mit Parametern auszulagern.
-- Die RAND()-Funktion ist per Definition nicht "deterministisch"  -sie gibt also bei jedem 
-- Aufruf einen potentiell anderen Rckgabewert zurck. Nicht deterministische Funktionsaufrufe 
-- innerhalb von Funktionen sind unter T-SQL aber verboten. --> Auslagern des Zufalls in den
-- Aufruf der Funktion, die dann "nur" noch dafr sorgt, dass das Ergebnis ganzzahlig und
-- im gewnschten Bereich ist...
CREATE OR ALTER FUNCTION [Lotto].[fncZufallszahl]
(
	  @Untergrenze		AS INTEGER
	, @Obergrenze		AS INTEGER
	, @Zufallszahl		AS FLOAT
)
RETURNS INTEGER
AS
BEGIN
	DECLARE @Rueckgabewert		AS INTEGER
	SELECT @Rueckgabewert		= FLOOR(@Zufallszahl * (@Obergrenze - @Untergrenze) + @Untergrenze)
	RETURN @Rueckgabewert
END
GO

DECLARE @Untergrenze	AS INTEGER
DECLARE @Obergrenze		AS INTEGER
SET @Untergrenze		= 1
SET @Obergrenze			= 50

SELECT [Lotto].[fncZufallszahl](@Untergrenze, @Obergrenze, RAND()) AS [Zufallswert]
GO




-- 5c. Es bietet sich an, diesen Quellcode in eine Funktion mit Parametern auszulagern.
-- Die RAND()-Funktion ist per Definition nicht "deterministisch"  -sie gibt also bei jedem 
-- Aufruf einen potentiell anderen Rckgabewert zurck. Nicht deterministische Funktionsaufrufe 
-- innerhalb von Funktionen sind unter T-SQL aber verboten. --> Auslagern des Zufalls in eine
-- Sicht, die dann in der Zufallsprozedur gejoint wreden kann
CREATE OR ALTER VIEW [Lotto].[vZufallswert]
AS
	SELECT        RAND() AS [Zufallswert]
GO



CREATE OR ALTER FUNCTION [Lotto].[fncZufallszahl2]
(
	  @Untergrenze		AS INTEGER
	, @Obergrenze		AS INTEGER
)
RETURNS INTEGER
AS
BEGIN
	DECLARE @Rueckgabewert		AS INTEGER
	SELECT @Rueckgabewert		= @Untergrenze 
									+ ROUND((SELECT TOP 1 [Zufallswert] 
												FROM [Lotto].[vZufallswert]) * (@Obergrenze - @Untergrenze), 0);
	RETURN @Rueckgabewert
END
GO

DECLARE @Untergrenze	AS INTEGER
DECLARE @Obergrenze		AS INTEGER
SET @Untergrenze		= 1
SET @Obergrenze			= 50

SELECT [Lotto].[fncZufallszahl2](@Untergrenze, @Obergrenze) As [Zufallswert]
GO



-- #########################################################
-- ### Variante 2: Eine einzelne Zufallszahl per NEWID() ###
-- #########################################################

DECLARE @Untergrenze	AS INTEGER
DECLARE @Obergrenze		AS INTEGER
SET @Untergrenze		= 1
SET @Obergrenze			= 50

SELECT ABS(CHECKSUM(NEWID()) % @Obergrenze - @Untergrenze) + @Untergrenze
GO




-- ##########################################################
-- ### Variante 3: Eine einzelne Zufallszahl per ORDER BY ###
-- ##########################################################

DECLARE @Untergrenze	AS INTEGER
DECLARE @Obergrenze		AS INTEGER
SET @Untergrenze		= 1
SET @Obergrenze			= 49

-- Bau einer CTE
;WITH [Einzelziehung] ( n ) AS (
        SELECT @Untergrenze UNION ALL
        SELECT @Untergrenze + n FROM [Einzelziehung] WHERE n < @Obergrenze )
SELECT
    TOP 1
	[Einzelziehung].n AS [Ziehung]
FROM [Einzelziehung]
ORDER BY NEWID()
GO



-- ##########################################################
-- ### beachte den Wirkbereich!                           ###
-- ##########################################################
;WITH Durchgang ( n ) AS (
        SELECT 1 UNION ALL
        SELECT 1 + n FROM Durchgang WHERE n < 5 )
SELECT
      CAST(((49+ 1) - 0) * RAND() + 1 AS TINYINT)	AS [ZufallszahlPerRAND]		-- hier wird der Zufall generiert
    , (ABS(CHECKSUM(NEWID())) % 49 + 1)				AS [ZufallszahlPerNewID]	-- hier wird der Zufall generiert
FROM Durchgang
GO

